package com.streambaase.liveview.cluster.monitor;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentSkipListSet;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.atomic.AtomicBoolean;

import org.slf4j.Logger;

import com.streambase.liveview.client.AbstractQueryListener;
import com.streambase.liveview.client.BeginSnapshotEvent;
import com.streambase.liveview.client.EndDeleteEvent;
import com.streambase.liveview.client.EndSnapshotEvent;
import com.streambase.liveview.client.LiveResult;
import com.streambase.liveview.client.LiveViewConnection;
import com.streambase.liveview.client.LiveViewConnectionFactory;
import com.streambase.liveview.client.LiveViewException;
import com.streambase.liveview.client.LiveViewExceptionType;
import com.streambase.liveview.client.LiveViewQueryType;
import com.streambase.liveview.client.QueryClosedEvent;
import com.streambase.liveview.client.QueryConfig;
import com.streambase.liveview.client.QueryExceptionEvent;
import com.streambase.liveview.client.Table;
import com.streambase.liveview.client.TablePublisher;
import com.streambase.liveview.client.TupleAddedEvent;
import com.streambase.liveview.client.TupleRemovedEvent;
import com.streambase.liveview.client.TupleUpdatedEvent;
import com.streambase.liveview.client.internal.LiveViewConstants;
import com.streambase.sb.Schema;
import com.streambase.sb.Schema.Field;
import com.streambase.sb.StreamBaseException;
import com.streambase.sb.Timestamp;
import com.streambase.sb.Tuple;
import com.streambase.sb.TupleException;
import com.streambase.sb.adapter.OutputAdapter;
import com.streambase.sb.adapter.common.StatusPort;
import com.streambase.sb.operator.Operator;
import com.streambase.sb.operator.Parameterizable;
import com.streambase.sb.operator.PortCounts;
import com.streambase.sb.operator.TypecheckException;
import com.streambase.sb.util.Msg;
import com.streambase.sb.util.Util;

/**

 */
public class GetLVSystemTables extends OutputAdapter implements Parameterizable, Runnable {

	public static final long serialVersionUID = 1426012544910L;
	private Logger logger;
    private LogLevel logLevel;
    
	private String displayName = "Get LiveView System Tables";
    public final static String CLIENT_INFO_TAG="?" + LiveViewConstants.CLIENT_INFO;  
	
	private static final int CONTROL_PORT_NUMBER=0;
	private static final int PUBLISHSCHEMAS_PORT_NUMBER=1;
	
	private static final String COMMAND_FIELD_NAME="Command";
	private static final String SERVERNAME_FIELD_NAME="ServerName";
	private static final String APPENDER_FIELD_NAME="AppenderName";
	private static final String SUMMARYTIME_FIELD_NAME="SummaryTime";
	private static final String URI_FIELD_NAME="URI";
	
	private static final String CQSDELETE_NAME="CQSDelete";
	private static final String EVENTTIME_NAME="EventTime";
		
	private static final int STATUS_OUTPUT_PORT_NUMBER=0;
	private static final int LIVEVIEWSTAISTICS_OUTPUT_PORT_NUMBER=1;
	
	private static final int NUM_INPUT_PORTS=2;
	private int NumOutputPorts=2;
	private String REMOTE_SERVER_LOG_TABLE_NAME=Util.getSystemProperty("liveview.cluster.monitor.remote.serverlog.name", "LVServerLog");
	private String CLUSTER_LOG_TABLE_NAME="CServerLog";
	private static final String CQUERY_SUMMARY_NAME="CQuerySummary";
	private static final String SERVERUP_NAME="ServerUp";
	private final int ROW_LIMIT=Util.getIntSystemProperty("liveview.cluster.monitor.row.limit", 30000);
	private static final String publishNode=Util.getSystemProperty("liveview.cluster.monitor.node", "localhost");
	private static final int publisherPort = Util.getIntSystemProperty("liveview.port", 10080);
	private static final String monitorName="LiveViewClusterMonitor";
	private static final String publisherURI=Msg.format("{0}{1}={2}", String.format("lv://%s:%s", publishNode, publisherPort), CLIENT_INFO_TAG, monitorName);

	private final boolean REMOVE_REFERENCES_ON_CONNECT=Util.getBooleanSystemProperty("liveview.cluster.monitor.remove.ref.on.connect", true);
	private final boolean REMOVE_LOGENTRIES_ON_CONNECT=Util.getBooleanSystemProperty("liveview.cluster.monitor.remove.logentries.on.connect", false);
	
	private Schema controlInputSchema=null;
	private Schema tableDefinitionSchema=null;
	private StatusPort statusPort;
	
	protected Map<String, ServerDescription> serverMap=new ConcurrentHashMap<String, ServerDescription>();
	protected Map<String, LiveViewConnection> clientMap=new ConcurrentHashMap<String, LiveViewConnection>();
	
	protected Set<ServerDescription> serversToConnectTo = new ConcurrentSkipListSet<ServerDescription>(new SDComparitor());
	protected Set<ServerDescription> serversCurrentlyConnected = new ConcurrentSkipListSet<ServerDescription>(new SDComparitor());

	private LiveViewConnection localClient=null;
	private TablePublisher cQuerySummaryPublisher=null;
	private AtomicBoolean done=new AtomicBoolean(false);
	
	protected Map<String, Integer> tableToPortMap = new HashMap<String, Integer>();
	protected Map<String, Schema> tableSchemasMap = new HashMap<String, Schema>();
	
	// Class to hold information about a server
	public class ServerDescription {
		public final String ServerName; 
		public final String URI;
		public AtomicBoolean shouldReconnect=new AtomicBoolean(true);
		public AtomicBoolean exceptionWarned=new AtomicBoolean(false);
		public int retryCount=0;
		
		public ServerDescription(String ServerName, String URI) {
			this.ServerName=ServerName;
			this.URI=URI;
		}
	}
	
	private class SDComparitor implements Comparator<ServerDescription> {
		@Override
		public int compare(ServerDescription arg0, ServerDescription arg1) {
			return arg0.ServerName.compareTo(arg1.ServerName) ;
		}
	}
			
	/**
	 * The constructor is called when the Adapter instance is created, but before the Adapter 
	 * is connected to the StreamBase application.
	 */
	public GetLVSystemTables() {
		super();
		logger = getLogger();
		setPortHints(NUM_INPUT_PORTS, NumOutputPorts);
		setDisplayName(displayName);
		setShortDisplayName(this.getClass().getSimpleName());
		setLogLevel(LogLevel.INFO);
	}

	/**
	 * Typecheck this adapter.
	 */
	public void typecheck() throws TypecheckException {
		// TODO Make sure all properties have valid values.
		
		requireInputPortCount(NUM_INPUT_PORTS);
        controlInputSchema = getInputSchema(CONTROL_PORT_NUMBER);
        tableDefinitionSchema = getInputSchema(PUBLISHSCHEMAS_PORT_NUMBER);
        
        statusPort = new StatusPort(STATUS_OUTPUT_PORT_NUMBER, this, logger, null);
        setOutputSchema(STATUS_OUTPUT_PORT_NUMBER, statusPort.getSchema());

        List<String> tableNames=new ArrayList<String>();
        
        if (NumOutputPorts-1 != tableDefinitionSchema.getFieldCount()) {
        	throw new TypecheckException("The configured output port count must match the PublishSchemasIn schema");
        }
        
        int portCount=1;
        for (Field f : tableDefinitionSchema ) {
        	tableNames.add(f.getName());
        	Schema tableSchema=new Schema(null, f.getSchema().getFields());
        	tableSchemasMap.put(f.getName(), tableSchema);
        	setOutputSchema(portCount, tableSchema);
        	tableToPortMap.put(f.getName(), portCount++);
        } 
       
	}
	
    public PortCounts getPortCounts() throws TypecheckException {
    	return new PortCounts(NUM_INPUT_PORTS, NumOutputPorts);
    }

	/**
	 * Initialize the adapter. 
	 */
	public void init() throws StreamBaseException {
		super.init();
		
		// Register the object so it will be run as a thread managed by StreamBase.
		registerRunnable(this, true);

        try {
            Operator.setLogLevel(logger, getLogLevel());
        } catch (StreamBaseException e) {
            logger.warn(Msg.format("{0}: Failed to set the log level to {1}, most likely because a logging backend other than the default is being used. Current StreamBase log level will be used instead.",
                                     getName(), getLogLevel()));
        }
        
    	try {
    		localClient=LiveViewConnectionFactory.getConnection(publisherURI);
    		cQuerySummaryPublisher=localClient.getTable(CQUERY_SUMMARY_NAME).getTablePublisher(monitorName);
    	} catch (Throwable t) {
    		logger.warn(Msg.format("Failed to connect to monitor LDM: {0}", publisherURI), t);
    	}
    	
	}

	/**
	 * Shutdown adapter
	 */
	public void shutdown() {
		done.set(true);
		
		for (ServerDescription sd : serversCurrentlyConnected) {
			LiveViewConnection client=clientMap.remove(sd.ServerName);
			if (client!=null) {
				// stop the flow from the server
				client.close();
			}
		}
		serversCurrentlyConnected.clear();
		
		if (localClient !=null) {
			localClient.close();
			localClient=null;
		}
	}
	
	public void processTuple(int inputPort, Tuple tuple) throws StreamBaseException {
		try {
			processTupleImpl(inputPort, tuple);
		} catch (Throwable t) {
			logger.warn(getName(), t);
		}
	}

	/**
	 * 
	 * @param inputPort the input port that the tuple is from (ports are zero based)
	 * @param tuple the tuple from the given input port
	 * @throws StreamBaseException Terminates the application. 
	
	 */
	public void processTupleImpl(int inputPort, Tuple tuple) throws StreamBaseException {
		
		switch (inputPort) {
		
		case CONTROL_PORT_NUMBER:
			String cmd=tuple.getString(COMMAND_FIELD_NAME);
			String serverName=tuple.getString(SERVERNAME_FIELD_NAME);
			
			if ("add".equalsIgnoreCase(cmd)) {
				if (serverMap.containsKey(serverName)) {
					logger.warn(Msg.format("Server name already exists: {0}", serverName));
					return;
				}
				if (tuple.isNull(URI_FIELD_NAME)) {
					logger.warn(Msg.format("URI can not be null for add commands: {0}", tuple.toString(true)));
					return;
				}
				ServerDescription sd=new ServerDescription(serverName, tuple.getString(URI_FIELD_NAME));
				setCQuerySummaryState(sd, false);
				logger.info(Msg.format("Adding {0} at {1}", sd.ServerName, sd.URI));
				serverMap.put(serverName, sd);
				serversToConnectTo.add(sd);
			} else if ("remove".equalsIgnoreCase(cmd)) {
				ServerDescription sd=serverMap.get(serverName);
				if (sd == null) {
					logger.warn(Msg.format("Can remove unknown server : {0}", serverName));
					return;
				}
				logger.info(Msg.format("Removing {0} at {1}", sd.ServerName, sd.URI));
				sd.shouldReconnect.set(false);
				removeServer(sd);
			} else {
				logger.warn(Msg.format("Unknown command: {0}", cmd));
			}
			break;
			
		case PUBLISHSCHEMAS_PORT_NUMBER:
			// should never happen
			logger.warn(Msg.format("Unexpected tuple on schema port: {0}", tuple));
			break;
		}
	}
	
	public void setCQuerySummaryState(ServerDescription sd, boolean up) {
		if (cQuerySummaryPublisher ==null) {
			// Got disconnected from monitor LDM. Not much to do.
			return;
		}
		Tuple t=cQuerySummaryPublisher.getSchema().createTuple();
		try {
			t.setString(SERVERNAME_FIELD_NAME, sd.ServerName);
			t.setTimestamp(SUMMARYTIME_FIELD_NAME, new Timestamp(new Date()));
			t.setBoolean(SERVERUP_NAME, up);
			cQuerySummaryPublisher.publish(t);
		} catch (Exception e) {
			logger.debug(getName(), e);
			cQuerySummaryPublisher=null;
			logger.warn(Msg.format("Error setting ServerUp for {0} state {1} exception: {2}", sd.ServerName, up, e.getMessage()));
		}
	}
	
	public void removeServer(ServerDescription sd) {
		
		serversCurrentlyConnected.remove(sd);
		
		LiveViewConnection client=clientMap.remove(sd.ServerName);
		if (client!=null) {
			// stop the flow from the server
			client.close();
		} else {
			// Someone else got there before us
			logger.debug(Msg.format("Didnt find client to close for {0}", sd.ServerName));
		}
		
		if (!sd.shouldReconnect.get()) {
			if (serverMap.remove(sd.ServerName) == null) {
				logger.info(Msg.format("Removing server not found in map: {0}", sd.ServerName));
			}
			serversToConnectTo.remove(sd);
			// expunge it from all tables
			removeLocalReferences(sd);
		} else {
			// must have died...
			setCQuerySummaryState(sd, false);
		}
	}
	
	public void removeLocalReferences(ServerDescription sd) {
		// expunge all references to this server from all tables
        for (Field f : tableDefinitionSchema ) {
        	String ctable=f.getName();

        	String nameField=SERVERNAME_FIELD_NAME;
			if (CLUSTER_LOG_TABLE_NAME.equals(ctable)) {
				if (!REMOVE_LOGENTRIES_ON_CONNECT) {
					continue;
				}
				nameField=APPENDER_FIELD_NAME;
			}

			QueryConfig qc = new QueryConfig().setTable(ctable).
					setPredicate(Msg.format("{0}==\"{1}\"", nameField, sd.ServerName)).
					setMaxResultSetSize(ROW_LIMIT).
					setQueryType(LiveViewQueryType.DELETE);

			DeleteListener dl= new DeleteListener();
			try {
				localClient.registerQuery(qc, dl);
				dl.waitForDone();
			} catch (Exception e) {
				logger.debug(null, e);
				logger.warn(Msg.format("Problem deleting server: {0} ({1})", sd.ServerName, e.getMessage()));
				continue;
			}
        }
        
		QueryConfig qc = new QueryConfig().setTable(CQUERY_SUMMARY_NAME).setPredicate(Msg.format("{0}==\"{1}\"", SERVERNAME_FIELD_NAME, sd.ServerName)).
				setMaxResultSetSize(Integer.MAX_VALUE).
				setQueryType(LiveViewQueryType.DELETE);
		try {
			localClient.deleteRows(qc);
		} catch (Exception e) {
			logger.debug(null, e);
			logger.warn(Msg.format("Problem deleting server: {0} ({1})", sd.ServerName, e.getMessage()));
		}
	}

	public int getPortCount() {
		return NumOutputPorts;
	}
	public void setPortCount(int p) {
		NumOutputPorts=p;
	}

	@Override
	public void run() {

		try {
	    	logger.info(Msg.format("GetLVSystemTables thread started for URI {0}", publisherURI));
	    	
			while (shouldRun() && !done.get()) {
				
				Iterator<ServerDescription> it=serversToConnectTo.iterator();
				if (it.hasNext()) {
					while (it.hasNext()) {
						ServerDescription sd=it.next();
						if (connectImp(sd)) {
							if (!registerQueries(sd)) {
								logger.info(Msg.format("Connected to {0} at {1}, but query registration failed", sd.ServerName, sd.URI));
							} else {
								it.remove();
								logger.info(Msg.format("Connected to {0} at {1}", sd.ServerName, sd.URI));
							}
						}
					}
				}

				if ((localClient == null) || !localClient.isConnected()) {
					localClient=null;
					cQuerySummaryPublisher=null;
					try {
						localClient=LiveViewConnectionFactory.getConnection(publisherURI);
					} catch (Throwable t) {
						logger.warn(Msg.format("Failed getting LDM monitor connection: {0}", publisherURI), t);
					}
				}
				
				if (cQuerySummaryPublisher == null) {
					if (localClient != null) {
						try {
							cQuerySummaryPublisher=localClient.getTable(CQUERY_SUMMARY_NAME).getTablePublisher(monitorName);
						} catch (Throwable t) {
							logger.warn("Problem getting local QuerySummaryPublisher", t);
						}
					}
				}
				Util.sleep(1000);
			}
		} finally {
			logger.info(getName(), "GetLVSystemTables thread exiting...");
		}
	}
	
	private boolean connectImp(ServerDescription sd) {
        logger.debug(Msg.format("Connecting to {0} at {1}", sd.ServerName, sd.URI));
        
        try {
            LiveViewConnection client= LiveViewConnectionFactory.getConnection(Msg.format("{0}{1}={2}", sd.URI, CLIENT_INFO_TAG, monitorName));
            sd.retryCount=0;
            clientMap.put(sd.ServerName, client);
            sd.exceptionWarned.set(false);
            serversCurrentlyConnected.add(sd);
            
            if (REMOVE_REFERENCES_ON_CONNECT) {
            	removeLocalReferences(sd);
            }
            
            setCQuerySummaryState(sd, true);
        } catch (Throwable t) {
        	sd.retryCount++;
        	serversCurrentlyConnected.remove(sd);
        	String msg=Msg.format("Failed connecting to {0} at {1}, retry count {2} ", sd.ServerName, sd.URI, sd.retryCount);
        	logger.debug(msg, t);
        	if ((sd.retryCount==1) || (sd.retryCount%10==0)) {
        		logger.info(msg +  t.getMessage());
        		// If we don't update occasionally this server will be removed from the FiveMinuteSummary
        		setCQuerySummaryState(sd, false);
        	}
	        return false;
        }

        return true;
	}
	
	private boolean registerQueries(ServerDescription sd) {
		LiveViewConnection client=clientMap.get(sd.ServerName);
		if (client == null) {
			logger.warn(Msg.format("Server {0} client not found", sd.ServerName));
			return false;
		}
        
        for (Field f : tableDefinitionSchema ) {
        	String ctable=f.getName();
        	if (CLUSTER_LOG_TABLE_NAME.equals(ctable)) {
        		// handle server log table specially 
        		continue;
        	}
        	
        	// this requires that the cluster tables name be exactly the system table names prepended with one character ("C")
        	String table=ctable.substring(1, ctable.length());
        	
			QueryConfig qc = new QueryConfig().setTable(table).setPredicate("true").
					setMaxResultSetSize(ROW_LIMIT).
					setQueryType(LiveViewQueryType.SNAPSHOT_AND_CONTINUOUS);
			
			try {
				client.registerQuery(qc, new LiveResult(new TablePublisherListener(this, ctable, sd)));
			} catch (Throwable t) {
				logger.debug(null, t);
				logger.warn(Msg.format("Problem connecting to table: {0} ({1})", table, t.getMessage()));
				closeClient(sd);
				return false;
			}
		}
        
        try {
			List<Table> tables=client.listTables();
			for (Table t : tables) {
				String n=t.getName();
				if (REMOTE_SERVER_LOG_TABLE_NAME.equals(n)) {
					QueryConfig qc = new QueryConfig().setTable(n).setPredicate("true").
							setMaxResultSetSize(ROW_LIMIT).
							setQueryType(LiveViewQueryType.SNAPSHOT_AND_CONTINUOUS);

						client.registerQuery(qc, new LiveResult(new TablePublisherListener(this, CLUSTER_LOG_TABLE_NAME, sd)));
				}
			}
		} catch (LiveViewException e) {
			logger.warn(Msg.format("{0} Unable to register server log, if any", sd.ServerName), e);
			closeClient(sd);
			return false;
		}
        return true;
	}
	
	private void closeClient(ServerDescription sd) {
		LiveViewConnection client=clientMap.get(sd.ServerName);
		if (client == null) {
			return;
		}
		
		client.close();
		clientMap.remove(sd.ServerName);
        serversCurrentlyConnected.remove(sd);
	}
	
	/*
	 * Listener used by all the queries to the remove LiveViews. Data received is published to the supplied table.
	 */
	private class TablePublisherListener extends AbstractQueryListener {

		private final String tableName;
        private final int port;
        private final Operator adapter;
        private Tuple ctableTuple;
        private final ServerDescription sd;
        private boolean addTime=false;
        private String nameField=SERVERNAME_FIELD_NAME;
        
		public TablePublisherListener (Operator adapter, String tableName, ServerDescription sd) throws LiveViewException {
			this.tableName = tableName;
			this.sd=sd;
			port=tableToPortMap.get(tableName);
			this.adapter=adapter;
			Schema cTableSchema=tableSchemasMap.get(tableName);
			ctableTuple=cTableSchema.createTuple();
			if (EVENTTIME_NAME.equals(cTableSchema.getField(1).getName())) {
				addTime=true;
			}
			
			if (CLUSTER_LOG_TABLE_NAME.equals(tableName)) {
				nameField=APPENDER_FIELD_NAME;
				// We want the remote time, not a local time
				addTime=false;
			}
		}
		
		private void copyTuple(Tuple t) {
			try {
				ctableTuple.clear();
				try {
					ctableTuple.setString(nameField, sd.ServerName);
					if (addTime) {
						ctableTuple.setTimestamp(EVENTTIME_NAME, new Timestamp(new Date()));
					}
				} catch (Exception e) {
					logger.warn(Msg.format("Unexpected exception setting name for server {0} table {1} ({2}", sd.ServerName, tableName, e));
				}
				
				for (Field f : t.getSchema().getFields()) {
					try {
						ctableTuple.setField(f.getName(), t.getField(f));
					} catch (TupleException e) {
						// BUGBUG no such field goes here.
					}
				}
			} catch (Throwable throwable) {
				logger.warn(Msg.format("Unexpected throwable for server {0} table {1} ({2}", sd.ServerName, tableName, throwable));
			}
		}
		
		@Override
		public void tupleAdded(TupleAddedEvent event) {
			try {
				if (serversCurrentlyConnected.contains(sd)) {
					copyTuple(event.getTuple());
					adapter.sendOutput(port, ctableTuple);
				}
			} catch (StreamBaseException e) {
				logger.warn(Msg.format("Unexpected exception in Tuple Add for server {0} table {1} ({2}", sd.ServerName, tableName, e));
			}
		}

		@Override
		public void tupleUpdated(TupleUpdatedEvent event) {
			try {
				if (serversCurrentlyConnected.contains(sd)) {
					copyTuple(event.getTuple());
					adapter.sendOutput(port, ctableTuple);
				}
			} catch (StreamBaseException e) {
				logger.warn(Msg.format("Unexpected exception in Tuple Update for server {0} table {1} ({2}", sd.ServerName, tableName, e));
			}
		}

		@Override
		public void tupleRemoved(TupleRemovedEvent event) {
			copyTuple(event.getTuple());
			try {
				ctableTuple.setBoolean(CQSDELETE_NAME, true);
				adapter.sendOutput(port, ctableTuple);
			} catch (StreamBaseException e) {
				logger.warn(Msg.format("Unexpected exception in Tuple Delete for server {0} table {1} ({2}", sd.ServerName, tableName, e));
			}
		}

		@Override
		public void snapshotBegin(BeginSnapshotEvent event) {
			
		}

		@Override
		public void snapshotEnd(EndSnapshotEvent event) {
				
		}

		@Override
		public void exceptionRaised(QueryExceptionEvent event) {
			if (done.get()) {
				// ignore death throws
				return;
			}

			if (event.getException().getErrorType()==LiveViewExceptionType.UPDATE_OVER_LIMIT) {
				logger.warn(Msg.format("Exception for server {0} on table {1} ({2}) - just killing this query, continuing all others", sd.ServerName, tableName, event.getException()));
				return;
			}
			if (!sd.exceptionWarned.getAndSet(true)) {
				logger.warn(Msg.format("Exception for server {0} on table {1} ({2})", sd.ServerName, tableName, event.getException()));
			}
				
			removeServer(sd);

			if (sd.shouldReconnect.get()) {
				serversToConnectTo.add(sd);
			}
		}

		@Override
		public void queryClosed(QueryClosedEvent event) {
			if (done.get()) {
				// ignore death throws
				return;
			}
			logger.info(Msg.format("Server {0} closed Table {1}", sd.ServerName, tableName));

			removeServer(sd);

			if (sd.shouldReconnect.get()) {
				serversToConnectTo.add(sd);
			}
		}
	}

	/*
	 * Delete listener so we can wait until all the rows are removed
	 */
	private class DeleteListener extends AbstractQueryListener {
		
		CountDownLatch done= new CountDownLatch(1);
		
		public DeleteListener() {
			
		}
		
		public void waitForDone() throws InterruptedException {
			done.await();
		}
		
	    @Override
	    public void deleteEnd(final EndDeleteEvent event) {
	    	done.countDown();
	    }
	    
		@Override
		public void exceptionRaised(final QueryExceptionEvent event) {
			done.countDown();
		}
	}
	
    public void setLogLevel(LogLevel logLevel) {
        this.logLevel = logLevel;
    }
    public LogLevel getLogLevel() {
        return this.logLevel;
    }

}
