package com.streambaase.liveview.cluster.monitor;

import java.beans.IntrospectionException;

import com.streambase.sb.operator.parameter.*;

/**
 * A BeanInfo class controls what properties are exposed, add 
 * metadata about properties (such as which properties are optional), and access 
 * special types of properties that can't be automatically derived via reflection. 
 * If a BeanInfo class is present, only the properties explicitly declared in
 * this class will be exposed by StreamBase.
 */
public class GetLVSystemTablesBeanInfo extends SBSimpleBeanInfo {

	private final static String PORT_COUNT_NAME="portCount";
    public static final String LOG_LEVEL_PROPERTY_NAME = "logLevel";
    
	/*
	 * The order of properties below determines the order they are displayed within
	 * the StreamBase Studio property view. 
	 */
	public SBPropertyDescriptor[] getPropertyDescriptorsChecked() throws IntrospectionException  {
		return new SBPropertyDescriptor[] {
				
                new SBPropertyDescriptor(PORT_COUNT_NAME, GetLVSystemTables.class).
                displayName("Output port count").
                description("The output port count needs to match the number of tables to publish to").
                optional(),
                new JavaEnumPropertyDescriptor<GetLVSystemTables.LogLevel>(LOG_LEVEL_PROPERTY_NAME, GetLVSystemTables.class, GetLVSystemTables.LogLevel.class)
                .displayName("Log Level")
                .description("Sets the log level used by the adapter. This setting operates independently of the top-level StreamBase log level value.")
                .optional(),
		};
	}

}
