#!/bin/bash

# This is a simple script that will run the lv-store-vacuum command on the store path supplied
# The output of the command are published to the CServerLog table

if [ -z $1 ] ; then
    echo "vacuum.sh: You must supply a store directory path"
    exit -1
fi
store_path=$1

noop=`lv-client --version`
if [ $? -ne 0 ] ; then
    echo "Problem with lv-client. Is it on the PATH?"
    exit -1
fi
liveview_uri="lv://localhost:50080"
table_name="CServerLog"

logger_name="LVInfoStore"
log_ts=`date "+%Y%m%d%H%M%S"`
vacuum_log="/tmp/vacuum-$log_ts.log"
appender_name="VacuumReporter"

# do the vacuum
lv-store-vacuum -d $store_path 2>&1 > $vacuum_log

# Now publish the results to the server log table
# use the epoc as the starting index
# this is OK for low volume writes that happen infrequently
index=`date +%s`
ts=`date "+%Y-%m-%d %H:%M:%S%z"`

# AppenderName,EventIndex,EventTime,LoggerName,Text
while read -r line
do
    echo $appender_name,$index,"\"$ts\"",$logger_name,"\"$line\"" | lv-client -u $liveview_uri publish $table_name
    ((index=$index+1))
done < "$vacuum_log"
